# TMA521 - Large scale optimization
# Spring 2013
# Project 2
# Simon Sigurdhsson

# SETS
param T_HORIZON;
set T_ALL_INTERVALS := 0..T_HORIZON;
param maxjobs;
set JOBS := 1..maxjobs;
set K_RESOURCES ordered;
set K_mach_RESOURCES ordered;
param n {JOBS};
set ACTIVE_I {j in JOBS} := setof {i in 1..n[j]}(i);

# PARAMETERS
param T_length_interval;
param lambda_mach {JOBS,K_mach_RESOURCES};
param r_disc {JOBS};
param a_disc {K_mach_RESOURCES};
param proc_time_disc {JOBS};
param p_postmach {JOBS};
param d_disc {JOBS};

param c {j in JOBS, u in T_ALL_INTERVALS} =
  ((u+p_postmach[j]) + max(0,u+p_postmach[j]-d_disc[j]));
param y_disc_solution{JOBS, JOBS, K_RESOURCES};
param z_disc_solution{JOBS, K_RESOURCES};
param t_disc_solution{JOBS};

param nPROP {K_mach_RESOURCES} default 0;
param propx {JOBS,T_ALL_INTERVALS,k in K_mach_RESOURCES,1..nPROP[k]} >= 0;
param c_prop {k in K_mach_RESOURCES,1..nPROP[k]};
param pi {JOBS} default 0;
param piz {K_mach_RESOURCES} default 1;

# Things in the data we "don't need".
set I_OP; param M; param w; set Q_PREC;
param q_follow{Q_PREC}; param v_jq{Q_PREC}; param v_disc_jq_ext{Q_PREC};
param v_mach_jq{Q_PREC}; param proc_time_mach{JOBS};
param proc_time{I_OP,JOBS}; param p_j_o_postmach_disc{JOBS};
param a{K_RESOURCES}; param resource_weight{K_RESOURCES}; param r_mach{JOBS};
param r{JOBS}; param d{JOBS}; param lambda{I_OP,JOBS,K_RESOURCES};

# VARIABLES
var slack >= 0;
var alpha {k in K_mach_RESOURCES, 1..nPROP[k]} >= 0;
var x {JOBS,K_mach_RESOURCES,T_ALL_INTERVALS} binary;
var bin {JOBS,K_mach_RESOURCES,T_ALL_INTERVALS} integer;

# OBJECTIVE FUNCTIONS
# Master program
minimize MP:
    sum {k in K_mach_RESOURCES, i in 1..nPROP[k]} (c_prop[k,i]*alpha[k,i]);
minimize A_MP:
    slack;

# Subproblem
minimize A_SUB {k in K_mach_RESOURCES}:
    sum {j in JOBS, u in 0..T_HORIZON}
      (0 - pi[j]*x[j,k,u])
    - piz[k];

minimize SUB {k in K_mach_RESOURCES}:
    sum {j in JOBS, u in 0..T_HORIZON}
      ((c[j,u] - pi[j])*x[j,k,u])
    - piz[k];

# CONSTRAINTS
# Master program
subject to MP_multi {j in JOBS}:
    sum {u in T_ALL_INTERVALS, k in K_mach_RESOURCES, i in 1..nPROP[k]}
      propx[j,u,k,i]*alpha[k,i] + slack = 1;
subject to MP_convex {k in K_mach_RESOURCES}:
    sum {i in 1..nPROP[k]} alpha[k,i] = 1;
subject to MP_bin {j in JOBS, k in K_mach_RESOURCES, u in T_ALL_INTERVALS}:
    sum {i in 1..nPROP[k]} propx[j,u,k,i]*alpha[k,i] = bin[j,k,u];

# Subproblem
subject to SUB_limit  {k in K_mach_RESOURCES, j in JOBS}:
    sum {u in T_ALL_INTERVALS} x[j,k,u] <= lambda_mach[j,k];
subject to SUB_occupy {k in K_mach_RESOURCES, u in T_ALL_INTERVALS}:
    sum {j in JOBS, nu in max(u-proc_time_disc[j]+1,0)..u} x[j,k,nu] <= 1;
subject to SUB_delay  {k in K_mach_RESOURCES, j in JOBS,
  u in 0..(max(r_disc[j],a_disc[k])-1)}:
    x[j,k,u] = 0;

# Phase III
param opt{JOBS} >= 0;
minimize MPIII:
    sum {j in JOBS, k in K_mach_RESOURCES, u in T_ALL_INTERVALS}
      c[j,u]*x[j,k,u];
subject to MPIII_match {j in JOBS}:
    sum {k in K_mach_RESOURCES, u in T_ALL_INTERVALS} x[j,k,u] = opt[j];

# Feasibiliy problem
var t {I_OP,JOBS} >= 0;
var z {I_OP,JOBS,K_RESOURCES} binary;
var y {I_OP,JOBS,I_OP,JOBS,K_RESOURCES} binary;
var s {JOBS} >= 0;
var h {JOBS} >= 0;
minimize F:
    sum {j in JOBS} (s[j]-0.001*t[1,j]+h[j] +
      sum {i in ACTIVE_I[j], k in K_RESOURCES} resource_weight[k]*z[i,j,k]);
subject to F_scheduled {j in JOBS,i in ACTIVE_I[j]}:
    sum {k in K_RESOURCES} z[i,j,k] = 1;
subject to F_flex {j in JOBS,i in ACTIVE_I[j], k in K_RESOURCES}:
    z[i,j,k] <= lambda[i,j,k];
subject to F_order1 {j in JOBS, q in JOBS, i in ACTIVE_I[j],
  p in ACTIVE_I[q], k in K_RESOURCES}:
    y[i,j,p,q,k] + y[p,q,i,j,k] <= if (i=p and j=q) then 2 else z[i,j,k];
subject to F_order2 {j in JOBS, q in JOBS, i in ACTIVE_I[j],
  p in ACTIVE_I[q], k in K_RESOURCES}:
    y[i,j,p,q,k] + y[p,q,i,j,k] + 1 >= if (i=p and j=q)  then 0
      else (z[i,j,k] + z[p,q,k]);
subject to F_earliest {j in JOBS, q in JOBS, i in ACTIVE_I[j],
  p in ACTIVE_I[q], k in K_RESOURCES}:
    t[i,j] + proc_time[i,j] - M*(1-y[i,j,p,q,k]) <=
      if (i=p and j=q) then M else t[p,q];
subject to F_sorder {j in JOBS, i in 1..(n[j]-1)}:
    t[i,j] + proc_time[i,j] + w <= t[i+1,j];
subject to F_interop {j_prec in Q_PREC}:
    s[j_prec] + v_jq[j_prec] <= t[1,q_follow[j_prec]];
subject to F_release {j in JOBS}:
    t[1,j] >= r[j];
subject to F_avail {j in JOBS, i in ACTIVE_I[j], k in K_RESOURCES}:
    t[i,j] >= a[k]*z[i,j,k];
subject to F_completion {j in JOBS}:
    s[j] = t[n[j],j] + proc_time[n[j],j];
subject to F_tardiness {j in JOBS}:
    h[j] >= s[j]-d[j];
subject to F_fsol1 {j in JOBS, q in JOBS, k in K_RESOURCES}:
    y[2,j,2,q,k] = y_disc_solution[j,q,k];
subject to F_fsol2 {j in JOBS, k in K_RESOURCES}:
    z[2,j,k] = z_disc_solution[j,k];
